from pydantic import BaseModel
from .env_defaults import current
from .user_config import user_config

class EnvConfig(BaseModel):
    """
    The environment configuration for the Positron CLI.
    Loaded like so: defaults <- config file <- env vars.
    """
    API_BASE: str = current.api_base
    SOCKET_IO_DOMAIN: str = current.ws_base
    AUTH0_DOMAIN: str = current.auth0_domain
    AUTH0_CLIENT_ID: str = current.auth0_client_id
    AUTH0_AUDIENCE: str =  current.auth0_audience

    API_CREATE_JOB: str = None
    API_GET_PRESIGNED: str = None
    API_START_JOB: str = None
    SOCKET_IO_PATH: str = None
    COMPRESSED_WS_NAME: str = None
    USER_AUTH_TOKEN: str = None 

    def __init__(self):
        super().__init__()

        # Override app dev defaults
        if user_config.user_auth_token:
            self.USER_AUTH_TOKEN = user_config.user_auth_token
        if user_config.backend_api_base_url:
            self.API_BASE = user_config.backend_api_base_url
        if user_config.backend_ws_base_url:
            self.SOCKET_IO_DOMAIN = user_config.backend_ws_base_url

        # Set derived env vars
        self.SOCKET_IO_PATH = '/backend/api/ws/socket.io'
        self.API_CREATE_JOB = f'{self.API_BASE}/create-job'
        self.API_GET_PRESIGNED = f'{self.API_BASE}/generate-presigned-url'
        self.API_START_JOB = f'{self.API_BASE}/start-job'
        self.COMPRESSED_WS_NAME = 'workspace.tar.gz'

env = EnvConfig()
"""
The environment configuration for the Positron CLI.
"""



import typer
from typing_extensions import Annotated
from positron_common.deploy import positron_deploy
from positron_common.config import get_job_config
from positron_common.cli_args import args

def run_job(
  config: Annotated[str, typer.Option(help='The path to the job configuration file')] = 'job_config.yaml',
  debug: Annotated[bool, typer.Option(help='Enable debug logging')] = False,
  stream_stdout: Annotated[bool, typer.Option(help='Stream the job\'s stdout back to your CLI')] = False
):
    """
    Run a Generic job in the Positron Cloud

    You need to define a job_config.yaml that defines the job you want to run.
    Make sure you run this script in the same directory as your file or provide the path to the job_config.yaml file.
    """
    print(f'Running a job with configuration file: {config}')

    args.init(debug=debug, stream_stdout=stream_stdout)
    job_config = get_job_config(config)
    positron_deploy(job_config)



import sys
import os
import subprocess
import requests
import tarfile
import time
from threading import Thread
from datetime import datetime, timezone
import boto3
import json
import re
from positron_common.config import parse_job_config

'''_______________________________________________________________________________________
    
    INITIALIZING CONFIGURATION AND CONSTANTS
    ______________________________________________________________________________________
'''
# Disable stdout buffering
os.environ['PYTHONUNBUFFERED'] = '1'
os.environ['POSITRON_CLOUD_ENVIRONMENT'] = '1'

# Define API endpoints
API_BASE = os.environ.get('API_ENDPOINT')
API_GET_JOB = f'{API_BASE}/get-job'
API_UPDATE_JOB = f'{API_BASE}/update-job'
API_JOB_LIFECYCLE = f'{API_BASE}/check-lifecycle'

# Define AWS S3 attached folders
working_dir = "/usr/src/job-execution"
results_dir = "/usr/src/job-controller/ws/result"
aws_region = os.getenv("REGION", "us-west-2")

# Define cross component enums
job_statuses = dict(
    pending="pending",
    uploading="uploading",
    in_queue="in_queue",
    launching="launching_pod",
    pulling_image="pulling_image",
    pulled_image="pulled_image",
    starting_container="starting_container",
    started_container="started_container",
    initializing="initializing",
    computing="computing",
    storing="storing",
    complete="complete",
    failed="failed",
    execution_error="execution_error",
    terminate="terminate_job",
    terminated="terminated"
)
log_types = dict(
    stdout="INFO",
    stderr="ERROR",
    debug="DEBUG"
)

# Cycle intervals
charge_interval = int(os.getenv('POSITRON_CHARGE_INTERVAL', 60))
stdout_log_interval = int(os.getenv('POSITRON_STDOUT_INTERVAL', 2))
check_termination_interval = int(os.getenv('POSITRON_CHECK_TERMINATION_INTERVAL', 10))

# Job management
running_job = None
kill_threads = False
job_id = sys.argv[2]
job = {}
stdout, stderr = [], []
final_status = job_statuses['complete']
return_code = 0

# Request header
headers = {"PositronJobId": job_id,
           "SystemAuthenticationKey": os.environ['SYSTEM_AUTHENTICATION_KEY']}

# Global clients
cloudwatch_client = boto3.client('logs', region_name=aws_region)


'''_______________________________________________________________________________________
    
    SUB MODULES
    ______________________________________________________________________________________
'''
# Get Job -----------------------------------------------------------
def get_job():
    cloud_log('Fetching job details')
    resp = try_request(requests.get, API_GET_JOB, headers=headers)
    global job
    job = resp.json()


# Unpack Workspace  -------------------------------------------------
def unpack_workspace():
    try:
        cloud_log('Unpacking workspace')
        with tarfile.open('ws/workspace.tar.gz') as tar:
            tar.extractall(working_dir)
    except Exception as e:
        cloud_log(e, log_types["debug"])
        cloud_log('Extracting workspace failed', log_types["stderr"])


# Run Job -----------------------------------------------------------
def run_job():
    cloud_log('Starting job')
    update_job(status=job_statuses["computing"])
    
    config_file = os.path.join(working_dir, "job_config.yaml")
    is_generic, commands, env_vars = parse_job_config(config_file)

    if is_generic:
        run_commands(commands, env_vars)
    else:
        run_decorator_job(env_vars)


def run_commands(commands, env_vars):
    cloud_log(f'Found commands to execute:\n{commands}')

    # extract individual commands
    invalid_commands = []
    items = re.split(r'&&|;|\n', commands)
    items = [item.strip() for item in items if item.strip()]

    # validate commands
    for command in items:
        result = subprocess.run(f'command -v {command}', shell=True, capture_output=True)
        if result.returncode != 0:
            invalid_commands.append(command)

    if len(invalid_commands) > 0:
        # there are commands that can not be executed, stop job execution
        cloud_log(f'Can not execute commands! Following commands are invalid: {", ".join(set(invalid_commands))}', log_types["stderr"])
        global final_status; final_status = job_statuses["execution_error"]
        return

    # build command and execute as separate process
    start_and_monitor_job(["/bin/bash", "-c", commands], env_vars)


def run_decorator_job(env_vars):
    # Construct the command
    command_base = "pip install -r requirements.txt &&"
    entry_point = job['entryPoint']

    if entry_point is None:
        cloud_log('Entry point is required for decorator jobs!', log_types['stderr'])
        global final_status; final_status = job_statuses["execution_error"]
        return

    meta = job.get('meta', {})
    job_arguments = meta.get('jobArguments', [])

    # Join job arguments with spaces
    arguments_string = ' '.join(job_arguments)

    # Construct the execution command
    execution_command = f"python {entry_point} {arguments_string}"

    # Log the execution command
    cloud_log(f"Installing Dependencies and running: {execution_command}")

    # Combine the base command with the execution command
    full_command = f"{command_base} {execution_command}"

    # Final command to be used
    command = ["sh", "-c", full_command]
    start_and_monitor_job(command, env_vars)


def start_and_monitor_job(command, env_vars):
    # Start job

    # Copy current environment variables and update with custom env_vars
    env = os.environ.copy()
    
    if env_vars is not None:
        env.update(env_vars)

    global running_job; running_job = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=1, text=True, cwd=working_dir, env=env)

    # Start parallel threads
    termination_thread = start_termination_thread()
    charge_thread = start_charging_thread()
    stdout_thread = Thread(target=logging_thread, args=(running_job.stdout, log_types["stdout"])); stdout_thread.start()
    stderr_thread = Thread(target=logging_thread, args=(running_job.stderr, log_types["stderr"])); stderr_thread.start()

    # Wait for the process to finish
    running_job.wait()

    # Terminate threads
    global kill_threads; kill_threads = True
    stdout_thread.join()
    stderr_thread.join()
    charge_thread.join()
    termination_thread.join()

    global return_code; return_code = running_job.returncode
    global final_status
    if final_status != job_statuses['terminated'] and return_code > 0: final_status = job_statuses["execution_error"]


# Upload results ----------------------------------------------------
def upload_result():
    try:
        cloud_log('Copying results')

        # Ensure the result directory exists
        os.makedirs(results_dir, exist_ok=True)
        # Copy workspace to result directory (uploads result to S3 bucket)
        subprocess.run(["cp", "-r", f"{working_dir}/.", results_dir], check=True)

        cloud_log('Results successfully copied')

    except subprocess.CalledProcessError as e:
        cloud_log(f'CalledProcessError: {e}', log_types["stderr"])
    except PermissionError as e:
        cloud_log(f'PermissionError: {e}', log_types["stderr"])
    except Exception as e:
        cloud_log(f'An error occurred: {e}', log_types["stderr"])


# Finish Job --------------------------------------------------------
def finish_job():
    # Collect outputs and finalize the job
    out_str = '\n'.join(stdout)
    err_str = '\n'.join(stderr)

    tokens_used = job['tokensUsed']

    if tokens_used:
        tokens_used_msg = f'Total tokens used for processing job: {tokens_used}'
        cloud_log(tokens_used_msg)
        out_str += tokens_used_msg

    update_job(status=final_status, end_date=datetime.now(timezone.utc), output_log=out_str, error_log=err_str)
    
    # Shared backend checks for this signal to terminate socket connection
    cloud_log("POSITRON_SIGNAL_EOS")


'''_______________________________________________________________________________________
    
    UTILS
    ______________________________________________________________________________________
'''
# Check for termination ---------------------------------------------
def is_job_active():
    resp = try_request(requests.get, API_GET_JOB, headers=headers)
    db_job = resp.json()
    inactive_statuses = [
        job_statuses["complete"],
        job_statuses["failed"],
        job_statuses["execution_error"],
        job_statuses["terminate"],
        job_statuses["terminated"],
    ]

    if db_job['status'] not in inactive_statuses:
        return True

    global final_status
    if db_job['status'] == job_statuses["terminate"]:
        final_status = job_statuses["terminated"]
        cloud_log("The job has been terminated by the user.")
    else:
        final_status = db_job['status']
        cloud_log(f"The job is not active anymore (status: {db_job['status']})")

    return False


# Update Job --------------------------------------------------------
def update_job(status, start_date=None, end_date=None, output_log=None, error_log=None):
    cloud_log(f'Updating status: {status}')
    body={
        "status": status,
        "startDate": start_date.isoformat() if start_date is not None else None,
        "endDate": end_date.isoformat() if end_date is not None else None,
        "outputLog": output_log,
        "errorLog": error_log
    }
    # filter out the items where the value is None
    body = {key:value for key, value in body.items() if value is not None}
    try_request(requests.post, API_UPDATE_JOB, headers=headers, json_data=body)

# Charging thread ---------------------------------------------------
def start_charging_thread():
    cloud_log('Start charging thread', log_types["debug"])
    def charge_thread():
        while not kill_threads:
            res = try_request(requests.post, API_JOB_LIFECYCLE, headers=headers)
            if res is None:
                # stop looping, job is forced to terminate by try_request
                break
            j = res.json()
            if not j['succeeded']:
                # unable to validate job lifecycle due to known reason (e.g. insufficient funds, time/token limit exceeded)
                cloud_log('Job lifecycle error!', log_types["stderr"])
                cloud_log(j['message'], log_types["stderr"])

            # @Todo: discuss if job should be charged for every computational minute that is started, i.e. if termination occurs
            # between two charge calls, should we charge for the minute that was started but not completed?
            time.sleep(charge_interval) # @Todo: if these intervals are too long, termination of the job can take long
    ct = Thread(target=charge_thread)
    ct.start()
    return ct


# Termination thread ------------------------------------------------
def start_termination_thread():
    cloud_log('Starting termination thread', log_types["debug"])
    def termination_thread():
        while not kill_threads:
            res = try_request(requests.get, API_GET_JOB, headers=headers)
            if res is None:
                # stop looping, job is forced to terminate by try_request
                break
            j = res.json()
            if j['status'] == job_statuses['terminate']:
                cloud_log('Terminating job')
                terminate_running_job()
                global final_status; final_status = job_statuses["terminated"]
                break
            time.sleep(check_termination_interval)

    tt = Thread(target=termination_thread)
    tt.start()
    return tt


# Logging thread ----------------------------------------------------
def logging_thread(pipe, level):
    try: 
        with pipe:
            for line in pipe:
                if line.strip():
                    cloud_log(line.rstrip(), level)
    except:
        cloud_log(f'{level} logging pipe closed!')


# Terminate running job ---------------------------------------------
def terminate_running_job():
    global running_job
    # Signal termination
    running_job.stdout.close()
    running_job.stderr.close()
    running_job.kill()


# Print logs and send to cloud watch --------------------------------
def cloud_log(message: str, level = log_types["stdout"]):
    print(level, message)
    
    if level == log_types['stderr']: stderr.append(message)
    else: stdout.append(message)
    
    sys.stdout.flush()
    timestamp = int(round(time.time() * 1000))
    message = {'timestamp': timestamp, 'log_level': level, 'message': message, 'app_name': 'positron_wrapper'}
    log_events = [{
        'timestamp': timestamp,
        'message': json.dumps(message),
    }]

    log_group_name = os.getenv('AWS_JOB_LOG_GROUP_NAME')
    log_stream_name = f'positron-job-{job_id}'

    try:
        cloudwatch_client.put_log_events(
            logGroupName=log_group_name,
            logStreamName=log_stream_name,
            logEvents=log_events
        )
    except Exception as e:
        print(f'Failed to submit logs: {e}')


# Retry API requests ------------------------------------------------
def try_request(request_func, url, retries=2, headers=None, json_data=None):
    for attempt in range(retries):
        try:
            response = request_func(url, headers=headers, json=json_data)
            response.raise_for_status()  # Raise an error for 4xx and 5xx status codes
            return response
        except requests.RequestException as e:
            cloud_log(f"Attempt {attempt + 1} failed: {e}", log_types["debug"])
            if attempt < retries - 1:
                cloud_log("Retrying...", log_types["debug"])
                time.sleep(2)  # Adding a delay before retrying
            else:
                cloud_log("Max retries reached, terminating job.", log_types["debug"])
                terminate_running_job()
                return None


'''_______________________________________________________________________________________
    
    RUN!
    ______________________________________________________________________________________
'''
def run():
    try:
        # 1. Get the job details
        get_job()

        if is_job_active():
            # 2. Update status to initializing
            update_job(status=job_statuses["initializing"], start_date=datetime.now(timezone.utc))

            # 3. Unpack workspace tar
            if is_job_active():
                unpack_workspace()

                # 4. Run the job (including dependency install)
                if is_job_active():
                    run_job()

                    # 5. Update status to Storing
                    update_job(status=job_statuses['storing'])

                    # 6. Upload results
                    upload_result()

        # 7. Get the completed job details
        get_job()
        # 8. Update status to success or error
        finish_job()

    except Exception as e:
        cloud_log(f'An exception occured: {str(e)}', log_types["stderr"])




from pydantic import BaseModel, field_validator
from typing import Dict, Optional
from typing_extensions import Self
import os
import yaml
import re
from .utils import PositronException

class PositronJob(BaseModel):
    """
    The Job details as defined in the `python_job` from the `job_config.yaml` file.
    """
    funding_group_id: Optional[str] = None
    environment_id: Optional[str] = None
    image: Optional[str] = None
    commands: Optional[str] = None
    entry_point: Optional[str] = None
    workspace_dir: Optional[str] = None
    max_tokens: Optional[str] = None
    max_time: Optional[str] = None
    env: Optional[Dict[str, str]] = None

    @field_validator('commands', mode='after')
    def ensure_non_empty(cls, commands):
        return commands if len(commands) else None

    @field_validator('env', mode='after')
    def ensure_env_is_dict(cls, v):
        if isinstance(v, dict):
            return v
        raise ValueError('env must be a dictionary')

    def validate(self) -> Self:
        errors = []
        if not self.funding_group_id:
            errors.append('funding_group is required')
        if not self.commands and not self.entry_point:
            errors.append('At least one of commands or entry_point must be provided')
        if self.env and not validate_env_vars(self.env):
            errors.append('At least one of the environment variables provided is invalid')
        if errors:
            raise PositronException(f'Invalid configuration. Errors: {errors}')

        return self

class PositronJobConfig(BaseModel):
    """
    The `job_config.yaml` schema class.
    """
    version: float
    python_job: PositronJob


def has_valid_prefix(key):
    """
    Check if the key starts with the required prefix.
    """
    return key.startswith("POSITRON_ENV_")

def is_valid_key_value(keyvalue):
    """
    Validate that the key-value contains only alphanumeric characters, dashes, and underscores, and has no spaces.
    """
    return bool(re.match(r'^[\w-]+$', keyvalue))

def validate_env_vars(env_dict):
    """
    Validate the environment variables from the given dictionary.
    """
    valid = True
    for key, value in env_dict.items():
        if not has_valid_prefix(key):
            print(f"Invalid key (missing prefix): {key}")
            valid = False
        if not is_valid_key_value(key):
            print(f"Invalid key (contains invalid characters or spaces): {key}")
            valid = False
        if not is_valid_key_value(value):
            print(f"Invalid value (contains invalid characters or spaces): {value}")
            valid = False
    return valid

def merge_config(base_config: PositronJob, override_config: PositronJob) -> PositronJob:
    """
    Makes it easy to merge decorator configs on top of the YAML config.
    """
    update_data = override_config.dict(exclude_unset=True)
    updated_config = base_config.copy(update=update_data)
    return updated_config

def get_job_config(config_path: str = 'job_config.yaml') -> Optional[PositronJob]:
    """
    Load the job configuration from the `job_config.yaml` file if it exists
    """
    if not os.path.exists(config_path):
        print('job_config.yaml file not found')
        return None

    try:
        with open(config_path, 'r') as job_config_file:
            job_config_dict = yaml.safe_load(job_config_file)
            job_config = PositronJobConfig(**job_config_dict)
            return job_config.python_job
    except Exception as e:
        print(f'Error loading job configuration! {str(e)}')

def parse_job_config(config_file):
    job_config_yaml = get_job_config(config_file)
    commands = getattr(job_config_yaml, 'commands', None)
    env_vars = getattr(job_config_yaml, 'env', None)
    # Determine if job is decorator job or generic
    return (commands is not None and len(commands) > 0), commands, env_vars


import typer
from .run_job import run

app = typer.Typer(help="Runs your job from the Positron Cloud environment.")

@app.command()
def hello():
    """
    Describes what you can do with the Positron Job Runner
    """
    print('Hello, I am the Positron Job Runner')
    print('Here is a list of thing I can help you with:')
    print('- Run a job in the Positron Cloud')

@app.command()
def run_job(job_id: str):
    """
    Run the job from inside a Positron container.

    Example usage:
    $ positron_job_runner run_job <job_id>
    """
    print('Running job in the Positron Cloud')
    print(f'Job ID: {job_id}')
    run()

if __name__ == "__main__":
    app()



import typer
import configparser
import requests
import time
import os
import webbrowser
from positron_common.deploy import positron_deploy
from positron_common.config import get_job_config
from positron_common.cli_args import args
from positron_common.utils import debug as print_d
from . import run_job, login

app = typer.Typer(help="A CLI tool to help you run your code in the Positron Cloud")

app.command()(run_job)
app.command()(login)

# TODO:
# @app.command()
# def setEnv(env: str = 'dev') # local, dev, alpha, beta
# update config.ini with proper urls

if __name__ == "__main__":
    app()



from functools import wraps
import os
import requests
import tarfile
import asyncio
import socketio
from colorama import init, Fore
import time
import signal
import json
import sys
import re
from .utils import debug, PositronException
from .enums import log_types
from .env_config import env
from .config import PositronJob
from .cli_args import args as cli_args
from .job_api.create_job import CreateJobBody

# Initialize colorama
init(autoreset=True)

# Cloud deployment definition
def positron_deploy(job_config: PositronJob):
    signal.signal(signal.SIGINT, handle_sigint)
    debug(env)

    try:
        # TODO: Move to PositronJob
        print('Validating sync parameters')
        validate_sync_parameters(job_config=job_config)
        job_config.validate()
        debug(job_config)

        image_name = job_config.image
        environment_id = job_config.environment_id
        # TODO: move to PositronJob
        workspace_dir = (job_config.workspace_dir if job_config.workspace_dir else get_default_workspace_dir())
        debug(f'Workspace directory: {workspace_dir}')
        entry_point = job_config.entry_point
        max_tokens = job_config.max_tokens
        max_time = job_config.max_time

        print(f'Environment: {environment_id if environment_id else "<Default>"}')
        print(f'Image: {image_name if image_name else "<Default>"}')

        print('Creating workspace tar file')
        create_workspace_tar(workspace_dir=workspace_dir)

        print('Creating new job')
        job = create_job(job_config=job_config)
        debug(json.dumps(job, indent=4))
        environment_id=job['environmentName']

        print('Fetching presigned url for upload')
        resp = get_presigned_url(job_id=job['id'])

        print('Uploading compressed workspace to Positron storage')
        upload_file(resp.get('url'), resp.get('fields'))

        print('Starting Job')
        start_job(job_id=job['id'])

        print("Your workspace has been uploaded and the job is in a processing queue. Please check your dashboard to follow your jobs status!")
        print(f"\tJob Name: {job['name']} ({job['id']})")
        print(f"\tFunding Group: {job['fundingGroupName']} ({job['fundingGroupId']})")
        print(f"\tEnvironment: {environment_id} ({job['environmentId']})")

        start_stdout_stream(job['id'])
        
    except PositronException as e:
        print(e)
    except Exception as e:
        print('An exception occurred. Please use --debug flag to get details')
        debug(e)


# May be able to move all this login into JobConfig
def validate_sync_parameters(job_config: PositronJob):
    def try_parse_max_tokens(tokens):
        if tokens is None:
            return None, None
        try:
            token_amount = tokens if isinstance(tokens, int) else int(tokens)
        except:
            return None, f'{tokens} is not an integer!'
        if token_amount > 0:
            return token_amount, None
        else:
            return None, 'It must be a positive integer!'


    def try_parse_max_time(time_string):
        if time_string is None:
            return None, None
        matches = re.search(r'^(\d+):(\d{2})$', time_string)
        if matches is None:
            return None, 'Format must be HH:MM'
        try:
            hours = int(matches.group(1))
            minutes = int(matches.group(2))
        except:
            return None, 'Values must be integers!'
        if minutes >= 60:
            return None, 'Invalid minutes! Must be 0 <= minutes < 60!'
        return hours * 60 + minutes, None
        
    valid = True
    max_tokens, error = try_parse_max_tokens(job_config.max_tokens)
    if error is not None:
        print(f'Invalid maximum amount of tokens! {error}')
        valid = False
    job_config.max_tokens = max_tokens
    
    max_time, error = try_parse_max_time(job_config.max_time)
    if error is not None:
        print(f'Invalid maximum execution time! {error}')
        valid = False
    job_config.max_time = max_time

    if not valid:
        raise PositronException(
            'Validation failed! Please check your decorator parameters!')


def create_workspace_tar(workspace_dir: str = None):
    # Use the context manager to handle opening and closing the tar file
    with tarfile.open(env.COMPRESSED_WS_NAME, 'w:gz') as tar:
        for root, dirs, files in os.walk(workspace_dir):
            for exclude_dir in ['.venv', '.git', '__pycache__']:
                try:
                    dirs.remove(exclude_dir)
                except ValueError:
                    # Ignore if the directory is not in the list
                    pass
            for file in files:
                full_path = os.path.join(root, file)
                arcname = os.path.relpath(full_path, start=workspace_dir)
                file_size = os.path.getsize(full_path)
                tar.add(full_path, arcname=arcname)
                debug(f"Added {full_path} as {arcname}, size: {file_size} bytes")


def create_job(job_config: PositronJob):
    debug(f'Calling: {env.API_CREATE_JOB}')
    debug(job_config)
    data = CreateJobBody.from_config(job_config)
    debug(data.json())
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN}
    response = requests.post(env.API_CREATE_JOB, headers=Headers, json=data.json())
    debug(response)
    if response.status_code != 200:
        body = response.json()
        if body.get('userFriendlyErrorMessage'):
            print(body.get('userFriendlyErrorMessage'))
            debug(json.dumps(body, indent=2))
            raise PositronException('Cannot create job. Please resolve the issue and try again.')
        raise PositronException(
            f'Job creation failed with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def get_presigned_url(job_id):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = env.API_GET_PRESIGNED + '?filename=' + env.COMPRESSED_WS_NAME
    debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    debug(response)
    if response.status_code != 200:
        raise PositronException(
            f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def upload_file(url, data):
    with open(env.COMPRESSED_WS_NAME, 'rb') as f:
        files = {'file': f}
        response = requests.post(url, data=data, files=files)
        if response.status_code != 204:
            raise PositronException(
                f'Upload failed with http code: {response.status_code} \n {response.text}')


def start_job(job_id):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    debug(f'Calling: {env.API_START_JOB}')
    response = requests.get(env.API_START_JOB, headers=Headers)
    debug(response)
    if response.status_code != 200:
        raise PositronException(
            f'Failed to start job with http code: {response.status_code} \n {response.text}')
    else:
        debug(response.json())
        return response.json()


def start_stdout_stream(job_id):
    if cli_args.stream_stdout:
        try:
            asyncio.get_event_loop().run_until_complete(start_stream(job_id))
        except Exception as e:
            print (e)


sio = socketio.AsyncClient()


@sio.event(namespace='/stdout-stream')
async def connect():
    print('Connected to stdout stream')


@sio.event(namespace='/stdout-stream')
async def message(message):
    try:
        log = json.loads(message)
        if log['log_level'] == log_types['stdout']:
            print(Fore.GREEN + log['message'])
        elif log['log_level'] == log_types['stderr']:
            print(Fore.RED + log['message'])
        elif log['log_level'] == log_types['debug']:
            print(Fore.BLUE + log['message'])            
    except:
        print(message)


@sio.event(namespace='/stdout-stream')
async def disconnect():
    print('Disconnected from stdout stream')


@sio.event(namespace='/stdout-stream')
async def error(err):
    print('An error occured in the streaming process')
    debug(err)


async def start_stream(job_id):
    custom_headers = {
        "PositronAuthToken": env.USER_AUTH_TOKEN,
        "PositronJobId": job_id
    }
    await sio.connect(env.SOCKET_IO_DOMAIN, headers=custom_headers, socketio_path=env.SOCKET_IO_PATH)
    await sio.wait()


def handle_sigint(signum, frame):
    print('Terminating gracefully...')
    time.sleep(5)
    exit(0)

def get_default_workspace_dir():
    cwd = os.getcwd()
    # decorator path calculation from file being run
    dirname = os.path.dirname(sys.argv[0])
    # TODO: cli and decorator calculate the workspace dir differently
    # think about moving this logic into those packages.
    # Run from current directory for generic jobs
    if (sys.argv[1] == 'run-job'):
        dirname = ''
    normalized_dir = os.path.normpath(dirname)
    absolute_path = os.path.join(cwd, normalized_dir)
    return os.path.normpath(absolute_path)



import sys

def debug(log):
    if sys.argv.__contains__('--debug'):
        print(f'DEBUG: {log}')

class PositronException(Exception):
    "Raised when decorator related errors are happening"
    pass

# Sentinel object for undefined
undefined = object()



from enum import Enum
from importlib.resources import files
from .utils import debug

class EnvType(Enum):
    DEV = "dev"
    ALPHA = "alpha"
    BETA = "beta"

class BuildEnvFile():
  """
  Plain text file that reflects which env this package was build for.
  This will drive the default values that can be overridden via config options.
  """
  value: EnvType = EnvType.DEV

  def __init__(self):
    self.value = self.get_type()

  def get_text(self):
    build_env_resource = files('positron_common').joinpath('build_env')
    try:
      return build_env_resource.read_text().strip()
    except Exception as e:
      debug(f'Error reading build_env file: {e}, defaulting to "{self.value.value}"')
      return self.value.value

  def get_type(self):
    try:
      return EnvType(self.get_text())
    except ValueError as e:
      print(f"There seems to be a problem with the build: {e}")
      print(f"Defaulting to {self.value.value}")
      return self.value

build_env: EnvType = BuildEnvFile().value
"""
The type of env this package was built for.
Used when selecting default values for the package.
"""



from pydantic import BaseModel
from typing import List, Optional

class PositronCLIArgs(BaseModel):
    """
    Positron CLI command line arguments.
    """
    is_init: bool = False
    local: bool = False
    deploy: bool = False
    stream_stdout: bool = False
    debug: bool = False
    job_args: Optional[List[str]] = None

    def init(self,
        local: bool = False,
        deploy: bool = False,
        stream_stdout: bool = False,
        debug: bool = False,
        job_args: Optional[List[str]] = None
    ):
        if self.is_init:
            raise ValueError('CLI Args already initialized')
        
        self.local = local
        self.deploy = deploy
        self.stream_stdout = stream_stdout
        self.debug = debug
        self.job_args = job_args
        self.is_init = True

#
# Export global (singleton)
#
args = PositronCLIArgs()
"""
Global CLI arguments singleton, make sure you call init() before using it.
"""



import yaml
from typing import Optional
from pydantic import BaseModel
import configparser
import os
from .utils import debug

DEFAULT_USER_CONFIG_PATH = os.path.expanduser('~/.positron/config.yaml')

class UserConfigFile(BaseModel):
    """
    The user configuration file for the Positron CLI.
    """
    user_auth_token: Optional[str] = None
    backend_api_base_url: Optional[str] = None
    backend_ws_base_url: Optional[str] = None

    def migrate(self):
        """
        Migrate an existing .ini file to .yaml, once.
        Leave old file in place so that switching to old and new versions works.
        """
        old_config_path = os.path.expanduser('~/.positron/config.ini')
        new_config_path = DEFAULT_USER_CONFIG_PATH
        if os.path.exists(new_config_path):
            # We've already migrated.
            return

        debug('config migration from ini to yaml required')
        old_config = configparser.ConfigParser()
        old_config.read(old_config_path)    
        if not old_config['DEFAULT']:
            debug('old config file is empty, nothing to migrate')
            self.write()
            return
        
        # write out new config from old
        new_config = UserConfigFile(
            user_auth_token=old_config['DEFAULT'].get('userauthtoken'),
            backend_api_base_url=old_config['DEFAULT'].get('POSITRON_API_BASE_URL'),
            backend_ws_base_url=old_config['DEFAULT'].get('POSITRON_WS_BASE_URL')
        )
        debug(f'previous configs: {new_config}')
        new_config.write()
        debug('Done migrating to new yaml config file!')

    
    def load_config(self):
        """
        Loads the user configuration from the user's home directory.
        """
        if not os.path.exists(DEFAULT_USER_CONFIG_PATH):
            debug(f'user config file not found: `{DEFAULT_USER_CONFIG_PATH}`')
            print('Could not find a user config file, please run `positron login` to set up your configuration.')
            return

        try:
            with open(DEFAULT_USER_CONFIG_PATH, 'r') as job_config_file:
                job_config_dict = yaml.safe_load(job_config_file)
                self.__init__(**job_config_dict)
        except Exception as e:
            print(f'Error loading job configuration! {str(e)}')
            print(f'Ignoring user config file: `{DEFAULT_USER_CONFIG_PATH}`')
    
    def write(self):
        """
        Write the user configuration to the user's home directory.
        """
        data = self.model_dump(exclude_none=True)
        os.makedirs(os.path.dirname(DEFAULT_USER_CONFIG_PATH), exist_ok=True)
        with open(DEFAULT_USER_CONFIG_PATH, 'w') as user_config_file:
            yaml.dump(data, user_config_file)

user_config = UserConfigFile()
"""
The user configuration file for the Positron CLI.
"""

if os.getenv('POSITRON_CLOUD_ENVIRONMENT', False):
    debug("Cloud environment detected, skipping config file load.")
else:
    debug('loading user config')
    user_config.migrate()
    user_config.load_config()



from functools import wraps
import sys
import os
import argparse
from positron_common.utils import debug
from positron_common.deploy import positron_deploy
from positron_common.config import PositronJob, get_job_config, merge_config
from positron_common.cli_args import args as cli_args

# Decorator definition
def positron_sync(**positron_parameters):

    # Parse command line arguments
    parser = argparse.ArgumentParser(description = "A decorator to handle deploying your code into the cloud")
    parser.add_argument('-l', '--local', action='store_true', help='Run your script locally', dest='local')
    parser.add_argument('--positron-deploy', action='store_true', help='Deploy your script into Positron Cloud', dest='deploy')
    parser.add_argument('--stream-stdout', action='store_true', help='Stream the stdout from Positron Cloud back to your cli', dest='stream_stdout')
    parser.add_argument('--debug', action='store_true', help='Get more detailed error messages', dest='debug')
    positron_args, job_args = parser.parse_known_args()

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            cli_args.init(
                local=positron_args.local,
                deploy=positron_args.deploy,
                stream_stdout=positron_args.stream_stdout,
                debug=positron_args.debug,
                job_args=job_args,
            )

            # get decorator parameters
            job_config_decorator = PositronJob(**positron_parameters)
            job_config = job_config_decorator

            # use job yaml as base if it exists
            job_config_yaml = get_job_config()
            if job_config_yaml:
                job_config = merge_config(job_config_yaml, job_config_decorator)

            # Need to ensure entry_point is set
            if not job_config.entry_point:
                job_config.entry_point = os.path.basename(sys.argv[0])

            # ensure valid config to run the job
            debug(job_config)
            job_config.validate()

            if cli_args.deploy:
                positron_deploy(job_config)
            elif cli_args.local or os.getenv('POSITRON_CLOUD_ENVIRONMENT', False):
                func(*args, **kwargs)
            else:
                parser.print_help()

        return wrapper
    return decorator



from typing import Optional, List
from pydantic import BaseModel, Field
from ..utils import undefined
from ..config import PositronJob
from ..cli_args import args as cli_args

class CreateJobBody(BaseModel):
    """
    Maps to the request body of the Create Job API
    """
    fundingGroupId: str
    imageName: Optional[str] = Field(default=undefined)
    environmentId: Optional[str] = Field(default=undefined)
    jobArguments: List[str] = Field(default=undefined)
    entryPoint: Optional[str] = Field(default=undefined)
    commands: Optional[str] = Field(default=undefined)
    maxTokens: Optional[int] = Field(default=undefined)
    maxMinutes: Optional[int] = Field(default=undefined)

    def json(self):
        """
        Enables dropping fields that were never set and should be treated as undefined
        """
        return {k: v for k, v in self.__dict__.items() if v is not undefined}

    @staticmethod
    def from_config(job_config: PositronJob):
      instance = CreateJobBody(
          fundingGroupId=job_config.funding_group_id,
          maxTokens=job_config.max_tokens,
          maxMinutes=job_config.max_time,
      )
      # determine if this is a generic or python job.
      if job_config.commands:
          instance.commands = job_config.commands
      elif job_config.entry_point:
          instance.entryPoint = job_config.entry_point
          instance.jobArguments = cli_args.job_args
      if job_config.image:
          instance.imageName = job_config.image
      if job_config.environment_id:
          instance.environmentId = job_config.environment_id
      return instance




import typer
from typing_extensions import Annotated
import requests
import time
import os
import webbrowser
from positron_common.env_config import env
from positron_common.cli_args import args
from positron_common.utils import debug as print_d
from positron_common.user_config import user_config

def login(
  debug: Annotated[bool, typer.Option(help='Enable debug logging')] = False,
):
    """
    Logs you in to your Positron account
    """
    args.init(debug=debug)

    # Get device code
    print('Requesting device code')
    device_code_payload = {
        'client_id': env.AUTH0_CLIENT_ID,
        'scope': 'openid profile',
        'audience': env.AUTH0_AUDIENCE
    }
    device_code_response = requests.post(f'https://{env.AUTH0_DOMAIN}/oauth/device/code', data=device_code_payload)
    if device_code_response.status_code != 200:
        print('Error generating device code')
        raise typer.Exit(code=1)

    device_code_data = device_code_response.json()

    # Redirect to login
    print('1. On your computer or mobile device navigate to: ', device_code_data['verification_uri_complete'])
    print('2. Enter the following code: ', device_code_data['user_code'])
    print('3. Complete the login process')
    print('')
    webbrowser.open(url=device_code_data['verification_uri_complete'], new=2, autoraise=True)

    # Wait for authentication
    access_token = wait_for_access_token(device_code_data)
    print_d(f'Access Token: {access_token}')

    # Get user token
    print('Requesting User Auth Token')
    auth_header = {
        'Authorization': f'Bearer {access_token}'
    }
    user_token_response = requests.get(f'{env.API_BASE}/get-user-auth-token', headers=auth_header)

    if user_token_response.status_code != 200:
        print(user_token_response.json())
        print('Error getting User Auth Token')
        raise typer.Exit(code=1)
    user_token_response_data = user_token_response.json()

    # Save user token
    print(f'Creating positron configuration at: {os.path.expanduser("~/.positron")}')
    save_user_token(user_token_response_data['userAuthToken'])

def wait_for_access_token(device_code_data):
    token_payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code_data['device_code'],
        'client_id': env.AUTH0_CLIENT_ID
    }
    authenticated = False
    while not authenticated:
        print('Checking if the user completed the flow...')
        token_response = requests.post(f'https://{env.AUTH0_DOMAIN}/oauth/token', data=token_payload)

        token_data = token_response.json()
        if token_response.status_code == 200:
            print('Authenticated!')
            authenticated = True
            return token_data['access_token']
        elif token_data['error'] not in ('authorization_pending', 'slow_down'):
            print(token_data['error_description'])
            raise typer.Exit(code=1)
        else:
            time.sleep(device_code_data['interval'])

def save_user_token(user_token):
    user_config.user_auth_token = user_token
    user_config.write()
    



from typing import Dict
from dataclasses import dataclass
from .build_env import build_env, EnvType
from .utils import debug

@dataclass
class EnvDefaults():
  api_base: str
  ws_base: str
  auth0_domain: str
  auth0_audience: str
  auth0_client_id: str

env_config: Dict[EnvType, EnvDefaults] = {
  EnvType.DEV: EnvDefaults(
    api_base='https://dev.positronsupercompute.com/backend/api',
    ws_base='wss://dev.positronsupercompute.com',
    auth0_domain='dev-k1t01pbanrr04itm.us.auth0.com',
    auth0_client_id='Mr3GW8Ub4e2bLaEvZ5o0XK5pGfEhtH3d',
    auth0_audience='https://localhost/positron/api',
  ),
  EnvType.ALPHA: EnvDefaults(
    api_base='https://alpha.positronsupercompute.com/backend/api',
    ws_base='wss://alpha.positronsupercompute.com',
    auth0_domain='dev-k1t01pbanrr04itm.us.auth0.com',
    auth0_client_id='Mr3GW8Ub4e2bLaEvZ5o0XK5pGfEhtH3d',
    auth0_audience='https://localhost/positron/api',
  ),
  EnvType.BETA: EnvDefaults(
    api_base='https://beta.positronsupercompute.com/backend/api',
    ws_base='wss://beta.positronsupercompute.com',
    auth0_domain='positron-beta.us.auth0.com',
    auth0_client_id='ZW0vio95rYfbHrN7kE3PoUXwmPloBw7e',
    auth0_audience='https://beta/positron/api',
  ),
}

current = env_config[EnvType.DEV]

# Set current based on build
if build_env:
  current = env_config[build_env]
  debug(f'Using build environment: "{build_env.value}" backend url: "{current.api_base}"')
